@@ -41,19 +41,24 @@ from micropipenv import HashMismatch
 from thoth.common import cwd
 from thoth.common import ThothAdviserIntegrationEnum
 from thoth.common import get_justification_link as jl
+from thamos.exceptions import ConfigurationError
 from thamos.exceptions import NoProjectDirError
 from thamos.exceptions import NoRuntimeEnvironmentError
+from thamos.exceptions import RequirementsFileError
 from thamos.config import config as configuration
 from thamos.lib import advise_here as thoth_advise_here
 from thamos.lib import add_requirements_to_project
 from thamos.lib import collect_support_information_dict
+from thamos.lib import check_runtime_environment_run
 from thamos.lib import get_log
 from thamos.lib import get_package_from_imported_packages
 from thamos.lib import get_status
+from thamos.lib import get_last_analysis_id
 from thamos.lib import install as thamos_install
 from thamos.lib import list_python_package_indexes
 from thamos.lib import list_python_environments
 from thamos.lib import list_thoth_container_images
+from thamos.lib import load_dot_env
 from thamos.lib import load_files
 from thamos.lib import provenance_check as thoth_provenance_check
 from thamos.lib import print_dependency_graph
@@ -85,6 +90,44 @@ _TABLE_COLS_ALIGN = {
 }
 
 
+_REPORT_TRANSLATION_TABLE_INDEXES = {
+    "url": "Python package index URL",
+    "verify_ssl": "Verify SSL",
+}
+
+
+_REPORT_TRANSLATION_TABLE_IMAGES = {
+    "cuda_version": "CUDA version",
+    "datetime": "Analysis datetime",
+    "env_image_name": "Image name",
+    "env_image_tag": "Image tag",
+    "environment_name": "Environment name",
+    "environment_type": "Environment type",
+    "image_sha": "Container image SHA",
+    "os_name": "OS name",
+    "os_version": "OS version",
+    "package_extract_document_id": "Container image analysis",
+    "python_version": "Python version",
+    "thoth_image_name": "Thoth container image name",
+    "thoth_image_version": "Thoth container image version",
+}
+
+
+_REPORT_TRANSLATION_TABLE_WHATPROVIDES = {
+    "package_import": "Package import",
+    "index_url": "Python package index URL",
+    "package_name": "Python package name",
+    "package_version": "Python package version",
+}
+
+
+_REPORT_TRANSLATION_TABLE_ENVIRONMENTS = {
+    "os_name": "Operating system name",
+    "os_version": "Operating system version",
+    "python_version": "Python version",
+}
+
+
 def handle_cli_exception(func: typing.Callable) -> typing.Callable:
     """Suppress exception in CLI if debug mode was not turned on."""
     # noqa
@@ -102,7 +145,14 @@ def handle_cli_exception(func: typing.Callable) -> typing.Callable:
     return wrapper
 
 
-def _print_report(report: list, json_output: bool = False, title: Optional[str] = None):
+def _print_report(
+    report: list,
+    *,
+    json_output: bool = False,
+    title: Optional[str] = None,
+    show_header: bool = True,
+    translation_table: Optional[Dict[str, str]] = None,
+):
     """Print reasoning to user."""
     if json_output:
         click.echo(json.dumps(report, sort_keys=True, indent=2))
@@ -110,7 +160,7 @@ def _print_report(report: list, json_output: bool = False, title: Optional[str]
 
     console = Console()
     table = Table(
-        show_header=True,
+        show_header=show_header,
         header_style="bold green",
         title=title,
         box=box.MINIMAL_DOUBLE_HEAD,
@@ -127,11 +177,11 @@ def _print_report(report: list, json_output: bool = False, title: Optional[str]
     # Remove fields that can be an array - these are addition details that are supressed from the table output.
     header = header - to_remove
 
+    translation_table = translation_table or {}
     header_list = sorted(header)
     for item in header_list:
-        table.add_column(
-            item.replace("_", " ").capitalize(), style="cyan", overflow="fold"
-        )
+        item = translation_table.get(item) or item.replace("_", " ").capitalize()
+        table.add_column(item, overflow="fold")
 
     for item in report:
         row = []
@@ -375,8 +425,18 @@ def _error_virtual_environment(virtualenv_path: str) -> None:
     help="Specify explicitly runtime environment to get recommendations for; "
     "defaults to the first entry in the configuration file.",
 )
+@click.option(
+    "--no-pedantic/--pedantic",
+    "-P",
+    is_flag=True,
+    show_default=True,
+    envvar="THAMOS_RUN_NO_PEDANTIC",
+    help="Do not abort running the application if the runtime environment used does not match configuration.",
+)
 def run(
-    command: typing.List[str], runtime_environment: Optional[str] = None
+    command: typing.List[str],
+    runtime_environment: Optional[str] = None,
+    no_pedantic: bool = False,
 ) -> None:  # noqa: D412
     """Run the command in virtual environment.
 
@@ -384,18 +444,57 @@ def run(
 
       thamos run ./app.py
 
-      thamos run --runtime-environment "testing" -- flask --help
+      thamos run --runtime-environment "stage" -- flask --help
+
+      thamos run --runtime-environment "testing" --no-pedantic -- train.py
     """
     virtualenv_path = configuration.get_virtualenv_path(runtime_environment)
     if virtualenv_path is None:
         _LOGGER.error("No virtual environment found for %r", runtime_environment)
         sys.exit(1)
 
+    overlays_dir = configuration.get_overlays_directory(runtime_environment)
+
+    if not no_pedantic:
+        check_runtime_environment_run(
+            configuration.get_runtime_environment(runtime_environment)
+        )
+
+        if configuration.requirements_format == "pipenv":
+            project = configuration.get_project(runtime_environment)
+
+            if not project.pipfile_lock:
+                raise RequirementsFileError(
+                    f"No Pipfile.lock found, lock your dependencies using `{sys.argv[0]} advise`"
+                )
+
+            if project.pipfile.hash() != project.pipfile_lock.meta.hash:
+                raise RequirementsFileError(
+                    f"Pipfile hash does not correspond to the lock file hash, use `{sys.argv[0]} advise` "
+                    f"to lock your dependencies"
+                )
+
+            _LOGGER.info("Pipfile hash corresponds to the hash stated in the lock file")
+        else:
+            _LOGGER.warning(
+                "Cannot verify requirements hash as Pipenv files are not used"
+            )
+
+    process_env = dict(os.environ)
+    if (
+        overlays_dir
+    ):  # Load any .env in the overlays directory or top root project directory.
+        process_env = {
+            **process_env,
+            **load_dot_env(os.path.join(overlays_dir, ".env")),
+        }
+
     python_path = os.path.join(virtualenv_path, "bin", "python3")
     _error_virtual_environment(virtualenv_path)
     # No magic here.
     cmd = [python_path, *command]
-    subprocess.run(cmd, universal_newlines=True, check=True)
+    _LOGGER.debug("Running application using %r", cmd)
+    subprocess.run(cmd, universal_newlines=True, check=True, env=process_env)
 
 
 @cli.command("venv")
@@ -418,7 +517,6 @@ def venv(ctx, runtime_environment: Optional[str] = None) -> None:
         ctx.exit(1)
     else:
         _error_virtual_environment(virtualenv_path)
-        print(virtualenv_path)
 
 
 @cli.command("purge")
@@ -579,7 +677,7 @@ def purge(
     metavar="KEY1=VALUE1,KEY2=VALUE2",
     default=None,
     show_default=True,
-    help="Labels used used to label the request.",
+    help="Labels used to label the request.",
 )
 @handle_cli_exception
 def advise(
@@ -632,7 +730,7 @@ def advise(
     # In CLI we always call to obtain only the best software stack (count is implicitly set to 1).
     results = thoth_advise_here(
         recommendation_type=recommendation_type,
-        src_path=configuration.config_path,
+        src_path=".",
         runtime_environment_name=runtime_environment,
         debug=debug,
         nowait=no_wait,
@@ -814,14 +912,25 @@ def provenance_check(
 @cli.command("log")
 @click.pass_context
 @click.argument("analysis_id", type=str, required=False, metavar="ANALYSIS_ID")
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment to which the given package should be added.",
+)
 @handle_cli_exception
-def log(analysis_id: typing.Optional[str] = None):
+def log(
+    analysis_id: typing.Optional[str] = None,
+    runtime_environment: typing.Optional[str] = None,
+):
     """Get log of running or finished analysis.
 
     If ANALYSIS_ID is not provided, the last request is used by default.
     """
     if not analysis_id:
-        with workdir():
+        with cwd(configuration.get_overlays_directory(runtime_environment)):
             log_str = get_log()
     else:
         log_str = get_log(analysis_id)
@@ -832,6 +941,14 @@ def log(analysis_id: typing.Optional[str] = None):
 @cli.command("status")
 @click.pass_context
 @click.argument("analysis_id", type=str, required=False, metavar="ANALYSIS_ID")
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment to which the given package should be added.",
+)
 @click.option(
     "--output-format",
     "-o",
@@ -841,7 +958,9 @@ def log(analysis_id: typing.Optional[str] = None):
 )
 @handle_cli_exception
 def status(
-    analysis_id: typing.Optional[str] = None, output_format: typing.Optional[str] = None
+    analysis_id: typing.Optional[str] = None,
+    output_format: typing.Optional[str] = None,
+    runtime_environment: typing.Optional[str] = None,
 ) -> None:  # noqa: D412
     """Get status of an analysis.
 
@@ -854,25 +973,17 @@ def status(
       thamos status "adviser-940101080006-110c392feb7cf6da"
     """
     if not analysis_id:
-        with workdir():
-            status_dict = get_status()
-    else:
-        status_dict = get_status(analysis_id)
+        with cwd(configuration.get_overlays_directory(runtime_environment)):
+            analysis_id = get_last_analysis_id()
+
+    status_dict = get_status(analysis_id)
 
     if not output_format or output_format == "table":
-        table = Table()
-
-        for key in status_dict.keys():
-            table.add_column(
-                key.replace("_", " ").capitalize(),
-                style="cyan",
-                overflow="fold",
-            )
-
-        table.add_row(*status_dict.values())
-
-        console = Console()
-        console.print(table, justify="center")
+        _print_report(
+            [status_dict],
+            json_output=False,
+            title=f"The current status for {analysis_id!r}",
+        )
         return
     elif output_format == "json":
         output = json.dumps(status_dict, indent=2)
@@ -895,9 +1006,18 @@ def status(
     type=bool,
     help="Collapse repeating sub-graphs in the output.",
 )
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment to which the given package should be added.",
+)
 @handle_cli_exception
 def graph(
     analysis_id: typing.Optional[str] = None,
+    runtime_environment: typing.Optional[str] = None,
     fold: bool = True,
 ) -> None:  # noqa: D412
     """Show dependency graph of resolved dependencies.
@@ -910,7 +1030,7 @@ def graph(
 
       thamos graph "adviser-940101080006-110c392feb7cf6da"
     """
-    with workdir():
+    with cwd(configuration.get_overlays_directory(runtime_environment)):
         printed = print_dependency_graph(analysis_id, fold=fold)
         if not printed:
             sys.exit(1)
@@ -1085,38 +1205,17 @@ def check(runtime_environment: Optional[str], output_format: str) -> None:  # no
         json.dump(result, sys.stdout, indent=2)
         sys.stdout.write("\n")
     elif output_format == "table":
-        table = Table()
+        _print_report(
+            result,
+            json_output=False,
+            title="Runtime environment and configuration check results",
+        )
 
-        header = set()
-        for item in result:
-            for key in item.keys():
-                header.add(key)
+    return_code = 1 if any(item.get("type") == "ERROR" for item in result) else 0
+    if return_code == 0:
+        _LOGGER.info("Configuration check passed!")
 
-        header_sorted = sorted(header)
-        for element in header_sorted:
-            table.add_column(
-                element.replace("_", " ").capitalize(),
-                style="cyan",
-                overflow="fold",
-            )
-
-        for item in result:
-            row = []
-            for key in header_sorted:
-                entry = item.get(key)
-                if not bool(int(os.getenv("THAMOS_NO_EMOJI", 0))) and isinstance(
-                    entry, str
-                ):
-                    entry = _EMOJI.get(entry, entry)
-
-                row.append(entry if entry is not None else "-")
-
-            table.add_row(*row)
-
-        console = Console()
-        console.print(table, justify="center")
-
-    sys.exit(1 if any(item.get("type") == "ERROR" for item in result) else 0)
+    sys.exit(return_code)
 
 
 @cli.command("images")
@@ -1128,15 +1227,106 @@ def check(runtime_environment: Optional[str], output_format: str) -> None:  # no
     default="table",
     help="Specify output format for the status report.",
 )
+@click.option(
+    "--os-name",
+    "-n",
+    type=str,
+    default=None,
+    metavar="OS_NAME",
+    help="Operating system name filter.",
+)
+@click.option(
+    "--os-version",
+    "-v",
+    type=str,
+    default=None,
+    metavar="OS_VERSION",
+    help="Operating system version filter.",
+)
+@click.option(
+    "--python-version",
+    "-p",
+    type=str,
+    default=None,
+    metavar="PY_VERSION",
+    help="Python interpreter version filter.",
+)
+@click.option(
+    "--cuda-version",
+    type=str,
+    default=None,
+    metavar="CUDA_VERSION",
+    help="CUDA version filter.",
+)
+@click.option(
+    "--image-name",
+    type=str,
+    default=None,
+    metavar="IMAGE_NAME",
+    help="Filter based on image name.",
+)
+@click.option(
+    "--library-name",
+    type=str,
+    default=None,
+    metavar="LIBRARY_NAME",
+    help="Filter based on library name.",
+)
+@click.option(
+    "--symbol",
+    type=str,
+    default=None,
+    metavar="SYMBOL",
+    help="Filter based on symbol.",
+)
+@click.option(
+    "--package-name",
+    type=str,
+    default=None,
+    metavar="PACKAGE_NAME",
+    help="Filter based on Python package name.",
+)
+@click.option(
+    "--rpm-package-name",
+    type=str,
+    default=None,
+    metavar="RPM_PACKAGE_NAME",
+    help="Filter based on RPM package name.",
+)
 @handle_cli_exception
-def images(output_format: str) -> None:  # noqa: D412
+def images(
+    output_format: Optional[str],
+    os_name: Optional[str],
+    os_version: Optional[str],
+    python_version: Optional[str],
+    cuda_version: Optional[str],
+    image_name: Optional[str],
+    library_name: Optional[str],
+    symbol: Optional[str],
+    package_name: Optional[str],
+    rpm_package_name: Optional[str],
+) -> None:  # noqa: D412
     """Check available Thoth container images.
 
     Examples:
 
       thamos images --output-format json
+
+      thamos images --os-name fedora --os-version 35 --python-version 3.9
+
+      thamos images --symbol GLIBC_FOO
     """
-    result = list_thoth_container_images()
+    result = list_thoth_container_images(
+        os_name=os_name,
+        os_version=os_version,
+        python_version=python_version,
+        cuda_version=cuda_version,
+        image_name=image_name,
+        library_name=library_name,
+        symbol=symbol,
+        package_name=package_name,
+        rpm_package_name=rpm_package_name,
+    )
 
     if output_format == "yaml":
         yaml.safe_dump({"s2i": result}, sys.stdout)
@@ -1144,38 +1334,22 @@ def images(output_format: str) -> None:  # noqa: D412
         json.dump({"s2i": result}, sys.stdout, indent=2)
         sys.stdout.write("\n")
     elif output_format == "table":
-        table = Table()
-
-        header = set()
         for item in result:
-            for key in item.keys():
-                header.add(key)
+            table_content = []
+            for key, value in item.items():
+                key = _REPORT_TRANSLATION_TABLE_IMAGES.get(
+                    key, key.replace("_", " ").capitalize()
+                )
+                key = Text.from_markup(key, style="green bold")
+                # Here, key value are ignored. This is a trick to have "table header" as the first column.
+                table_content.extend([{"key": key, "value": value}])
 
-        header_sorted = sorted(header)
-        header_sorted.append("info")
-        for item in header_sorted:
-            table.add_column(
-                item.replace("_", " ").capitalize(),
-                style="cyan",
-                overflow="fold",
+            _print_report(
+                table_content,
+                title=f"Container image {item.get('environment_name', 'UNKNOWN')}",
+                show_header=False,
             )
 
-        for item in result:
-            row = []
-            for key in header_sorted:
-                if key == "info":
-                    image_name = item.get("thoth_image_name")
-                    row.append(jl(image_name.rsplit("/", maxsplit=1)[-1]))
-                    continue
-
-                entry = item.get(key)
-                row.append(str(entry) if entry is not None else "-")
-
-            table.add_row(*row)
-
-        console = Console()
-        console.print(table, justify="center")
-
     sys.exit(1 if any(item.get("type") == "ERROR" for item in result) else 0)
 
 
@@ -1199,35 +1373,18 @@ def indexes(output_format: str) -> None:
         json.dump({"indexes": result}, sys.stdout, indent=2)
         sys.stdout.write("\n")
     elif output_format == "table":
-        table = Table()
-
-        header = set()
         for item in result:
-            for key in item.keys():
-                if key == "warehouse_api_url":
-                    continue
-                header.add(key)
+            item.pop("warehouse_api_url", None)
+            for key, value in item.items():
+                # Substitute booleans as rich does not interpret them.
+                if isinstance(value, bool):
+                    item[key] = str(value)
 
-        header_sorted = sorted(header)
-        for item in header_sorted:
-            table.add_column(
-                item.capitalize()
-                .replace("_", " ")
-                .replace("Url", "URL")
-                .replace("ssl", "SSL")
-                .replace("api", "API")
-            )
-
-        for item in result:
-            row = []
-            for key in header_sorted:
-                entry = item.get(key)
-                row.append(str(entry) if entry is not None else "-")
-
-            table.add_row(*row)
-
-        console = Console()
-        console.print(table, justify="center")
+        _print_report(
+            result,
+            title="Python package indexes available for consuming packages",
+            translation_table=_REPORT_TRANSLATION_TABLE_INDEXES,
+        )
 
     sys.exit(0)
 
@@ -1349,7 +1506,7 @@ def remove(
     configuration.save_project(project)
 
 
-@cli.command("support")
+@cli.command("support", short_help="Collect environment and configuration information.")
 @click.option(
     "--output",
     "-o",
@@ -1369,7 +1526,7 @@ def support(output: str) -> None:
             json.dump(info, output_file, sort_keys=True, indent=2)
 
 
-@cli.command("whatprovides")
+@cli.command("whatprovides", short_help="Check packages providing the given import.")
 @click.pass_context
 @click.option(
     "--output-format",
@@ -1390,7 +1547,11 @@ def whatprovides(import_name: str, output_format: str) -> None:  # noqa: D412
       thamos whatprovides sklearn.linear_model.*
     """
     _LOGGER.info("Returning information on package %r", import_name)
-    result = get_package_from_imported_packages(import_name)
+
+    result = get_package_from_imported_packages(import_name, raise_on_error=False)
+    if not result:
+        _LOGGER.error("No package providing %r found", import_name)
+        sys.exit(1)
 
     if output_format == "yaml":
         yaml.safe_dump({"packages": result}, sys.stdout)
@@ -1402,18 +1563,20 @@ def whatprovides(import_name: str, output_format: str) -> None:  # noqa: D412
             result,
             json_output=False,
             title=f"Packages for {import_name}",
+            translation_table=_REPORT_TRANSLATION_TABLE_WHATPROVIDES,
         )
 
     sys.exit(0)
 
 
-@cli.command("discover")
+@cli.command("discover", short_help="Discover packages used in the project.")
 @click.pass_context
 @click.option(
     "--src-path",
     "-s",
     type=str,
     default=".",
+    metavar="SRC_PATH",
     envvar="THAMOS_SOURCE_PATH",
     help="Specify path to consider to discover packages.",
 )
@@ -1429,14 +1592,25 @@ def whatprovides(import_name: str, output_format: str) -> None:  # noqa: D412
 def discover(
     runtime_environment: typing.Optional[str], src_path: str = "."
 ) -> None:  # noqa: D412
-    """Discover packages used in the project or in the file.
+    """Discover packages used in the project or in a file and add them to requirements.
+
+    If runtime environment is passed, requirements are added to requirements specific
+    to the given runtime environment. Otherwise, the default runtime environment is used.
 
     Examples:
 
       thamos discover
+
+      thamos discover --runtime-environment "production"
     """
     # Obtain list of imports using invectio and verify package from PyPI
-    verified_packages = get_verified_packages_from_static_analysis(src_path=src_path)
+    verified_packages = get_verified_packages_from_static_analysis(
+        src_path=src_path,
+        without_standard_imports=True,
+        without_builtin_imports=True,
+        without_builtins=True,
+        raise_on_error=False,
+    )
 
     # Update requirements files (Pipfile/Pipfile.lock) or requirements.txt (requirements logic)
     for package in verified_packages:
@@ -1477,7 +1651,78 @@ def environments_(output_format: str) -> None:  # noqa: D412
             environments,
             json_output=False,
             title="Python environments",
+            translation_table=_REPORT_TRANSLATION_TABLE_ENVIRONMENTS,
         )
 
 
+@cli.command("verify", short_help="Check lock file freshness.")
+@click.pass_context
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment from which the given package should be removed.",
+)
+@handle_cli_exception
+def verify(runtime_environment: typing.Optional[str]) -> None:  # noqa: D412
+    """Verify the hash in Pipfile.lock is up-to-date in runtime environments configured.
+
+    Examples:
+
+        thamos verify
+
+        thamos verify --runtime-environment "training"
+    """
+    if configuration.requirements_format != "pipenv":
+        _LOGGER.error(
+            "Cannot verify hash of the lock file as the project uses requirements format %r, please switch "
+            "to 'pipenv' to have the ability to check the lock file hash",
+            configuration.requirements_format,
+        )
+        sys.exit(2)
+
+    if runtime_environment:
+        runtime_environments = [
+            configuration.get_runtime_environment(runtime_environment)
+        ]
+    else:
+        runtime_environments = configuration.list_runtime_environments()
+
+    any_error = False
+    for runtime_environment_config in runtime_environments:
+        runtime_environment_name = runtime_environment_config["name"]
+
+        try:
+            project = configuration.get_project(runtime_environment_name)
+        except ConfigurationError:
+            _LOGGER.error(
+                "Failed to obtain configuration for runtime environment %r",
+                runtime_environment_name,
+            )
+            any_error = True
+            continue
+
+        if not project.pipfile_lock:
+            _LOGGER.error(
+                "No lock file found for runtime environment %r",
+                runtime_environment_name,
+            )
+            any_error = True
+            continue
+
+        if project.pipfile.hash() != project.pipfile_lock.meta.hash:
+            _LOGGER.error("Pipfile.lock is out-of-date. Run `thamos advise` to update.")
+            any_error = True
+            continue
+
+        _LOGGER.info(
+            "Pipfile.lock is up-to-date for runtime environment %r",
+            runtime_environment_name,
+        )
+
+    sys.exit(any_error is True)
+
+
 __name__ == "__main__" and cli()
