@@ -157,7 +157,6 @@ class _Configuration:
         self.explicit_host = None
         self.tls_verify = None
         self._api_url = None
-        self.config_path = None
 
     @property
     def api_url(self):
@@ -275,13 +274,11 @@ class _Configuration:
     def reset_config(self) -> None:
         """Discard loaded config in memory."""
         self._configuration = None
-        self.config_path = None
 
     def load_config(self, force: bool = False) -> None:
         """Load configuration from a file."""
         if not self._configuration and not force:
             with workdir(config.CONFIG_NAME):
-                self.config_path = os.getcwd()
                 self.load_config_from_file(config.CONFIG_NAME)
 
     def save_config(self, path: Optional[str] = None) -> None:
@@ -345,7 +342,9 @@ class _Configuration:
             platform=platform,
             requirements_format=requirements_format,
             base_image=base_image,
-            **cpu_info,
+            cpu_family=cpu_info["cpu_family"] or "null",
+            cpu_model=cpu_info["cpu_model"] or "null",
+            cpu_model_name=cpu_info["cpu_model_name"] or "CPU detection failed",
             **(dict(os.environ) if expand_env else {}),
         )
 
@@ -588,7 +587,7 @@ class _Configuration:
                 {
                     "type": "ERROR" if conf_cpu_family is not None else "WARNING",
                     "runtime_environment": runtime_environment_name,
-                    "message": f"CPU model stated in the configuration file ({conf_cpu_family}) does not match the "
+                    "message": f"CPU family stated in the configuration file ({conf_cpu_family}) does not match the "
                     f"one detected ({cpu_info.get('cpu_family')})",
                 }
             )
