@@ -17,9 +17,11 @@
 
 """Core parts of library for interacting with Thoth."""
 
+import datetime
 import logging
 import os
 import platform
+import random
 import sys
 import typing
 from itertools import chain
@@ -36,9 +38,9 @@ import requests
 import yaml
 import micropipenv
 from termcolor import colored
+from invectio import gather_library_usage
 from yaspin import yaspin
 from yaspin.spinners import Spinners
-from invectio import gather_library_usage
 from thoth.analyzer import run_command
 from thoth.python import Project
 from thoth.python import Constraints
@@ -63,11 +65,14 @@ from .swagger_client import EnvironmentsApi
 from .swagger_client import PythonPackagesApi
 from .swagger_client.models import AnalysisResultResponse
 from .config import config as thoth_config
-from .exceptions import UnknownAnalysisType
-from .exceptions import TimeoutError
+from .discover import discover_all
 from .exceptions import ApiError
 from .exceptions import NoDevRequirements
+from .exceptions import NoMatchingPackage
 from .exceptions import NoRequirementsFile
+from .exceptions import PedanticRunVerificationError
+from .exceptions import TimeoutError
+from .exceptions import UnknownAnalysisType
 
 from typing import Callable, Any, Union, Dict
 
@@ -118,7 +123,7 @@ def with_api_client(func: typing.Callable):
 
 
 def is_analysis_ready(analysis_id: str, *, verify_tls: bool = True) -> bool:
-    """Handle the the multiple response types available while asking for result of a anaylsis result."""
+    """Handle the multiple response types available while asking for result of an analysis."""
     config = Configuration()
     host = thoth_config.explicit_host
     if not host:
@@ -141,6 +146,31 @@ def is_analysis_ready(analysis_id: str, *, verify_tls: bool = True) -> bool:
     )
 
 
+def _get_spinner() -> object:
+    """Choose the spinner, wisely."""
+    today = datetime.datetime.now()
+    if today.month == 12 and today.day in (24, 25, 26):  # Christmas.
+        return Spinners.christmas
+    if (today.month == 1 and today.day == 1) or (
+        today.month == 12 and today.day == 31
+    ):  # End or start of the year.
+        return Spinners.earth
+    if today.month == 2 and today.day == 14:  # Valentine's day.
+        return Spinners.hearts
+    if today.month == 4 and today.day == 1:  # Fool's day.
+        return Spinners.smiley
+    if today.month == 4 and today.day == 17:  # Easter.
+        return Spinners.orangeBluePulse
+    if today.month == 10 and today.day == 31:  # Halloween.
+        return Spinners.monkey
+    if today.hour == 7:  # A morning run.
+        return Spinners.runner
+    if today.hour >= 22:  # Isn't it too late?
+        return Spinners.moon
+
+    return Spinners.clock if random.random() > 0.1 else Spinners.mindblown
+
+
 def _wait_for_analysis(
     status_func: Callable[..., Any],
     analysis_id: str,
@@ -156,7 +186,7 @@ def _wait_for_analysis(
 
     spinner = partial(
         yaspin,
-        Spinners.clock,
+        _get_spinner(),
         text=f"Waiting for response from Thoth (analysis: {analysis_id})...",
     )  # type: Union[Callable[..., Any], partial[Any]]
     if _LOGGER.getEffectiveLevel() == logging.DEBUG or bool(
@@ -223,7 +253,7 @@ def _note_last_analysis_id(analysis_id: str) -> None:
         _LOGGER.warning("Failed to write analysis id to a temporary file: %s", str(exc))
 
 
-def _get_last_analysis_id() -> str:
+def get_last_analysis_id() -> str:
     """Retrieve last analysis id from a temporary file."""
     try:
         with open(LAST_ANALYSIS_ID_FILE, "r") as analysis_id_file:
@@ -274,7 +304,13 @@ def _retrieve_analysis_result(
             sleep(_RETRY_ON_ERROR_SLEEP)
 
 
-def get_static_analysis(src_path: str = ".") -> typing.Optional[dict]:
+def get_static_analysis(
+    src_path: str = ".",
+    *,
+    without_standard_imports: bool = False,
+    without_builtin_imports: bool = False,
+    without_builtins: bool = False,
+) -> typing.Optional[dict]:
     """Get static analysis of files used in project."""
     # We are running in the root directory of project, use the root part for gathering static analysis.
     _LOGGER.info("Performing static analysis of sources to gather library usage")
@@ -282,6 +318,9 @@ def get_static_analysis(src_path: str = ".") -> typing.Optional[dict]:
         library_usage = gather_library_usage(
             src_path,
             ignore_errors=True,
+            without_standard_imports=without_standard_imports,
+            without_builtin_imports=without_builtin_imports,
+            without_builtins=without_builtins,
         )
     except FileNotFoundError:
         _LOGGER.warning("No library usage was aggregated - no Python sources found")
@@ -332,7 +371,7 @@ def advise_using_config(
     *,
     runtime_environment_name: typing.Optional[str] = None,
     constraints: typing.Optional[str] = None,
-    src_path: typing.Optional[str] = None,
+    src_path: str = ".",
     recommendation_type: str = None,
     dev: bool = False,
     no_static_analysis: bool = False,
@@ -362,7 +401,7 @@ def advise_using_config(
         pipfile_lock=pipfile_lock,
         constraints=constraints,
         recommendation_type=recommendation_type,
-        src_path=src_path if src_path is not None else thoth_config.config_path,
+        src_path=src_path,
         runtime_environment=thoth_config.get_runtime_environment(
             runtime_environment_name
         ),
@@ -560,8 +599,7 @@ def advise(
         return None
 
     _LOGGER.debug("Adviser check metadata: %r", response.metadata)
-
-    return response.result, response.result["error"]
+    return response.result.to_dict(), response.result.error
 
 
 def advise_here(
@@ -766,7 +804,8 @@ def provenance_check(
         return None
 
     _LOGGER.debug("Provenance check metadata: %r", response.metadata)
-    return response.result["report"], response.result["error"]
+    result = response.result.to_dict()
+    return result["report"], result["error"]
 
 
 def provenance_check_here(
@@ -927,7 +966,7 @@ def get_log(api_client: ApiClient, analysis_id: str = None):
     If analysis_id is not provided, its get from the last thamos call which stores it in a temporary file.
     """
     if not analysis_id:
-        analysis_id = _get_last_analysis_id()
+        analysis_id = get_last_analysis_id()
 
     if analysis_id.startswith("package-extract-"):
         api_instance = ImageAnalysisApi(
@@ -992,7 +1031,7 @@ def get_status(api_client: ApiClient, analysis_id: typing.Optional[str] = None):
     If analysis_id is not provided, its get from the last thamos call which stores it in a temporary file.
     """
     if not analysis_id:
-        analysis_id = _get_last_analysis_id()
+        analysis_id = get_last_analysis_id()
 
     if analysis_id.startswith("package-extract-"):
         api_instance = ImageAnalysisApi(
@@ -1019,7 +1058,7 @@ def get_analysis_results(
 ):
     """Get the analysis result from a given id."""
     if not analysis_id:
-        analysis_id = _get_last_analysis_id()
+        analysis_id = get_last_analysis_id()
 
     if analysis_id.startswith("package-extract-"):
         api_instance = ImageAnalysisApi(
@@ -1039,7 +1078,7 @@ def get_analysis_results(
         api_instance = AdviseApi(api_client)
         method = api_instance.get_advise_python  # type: ignore
         response = _retrieve_analysis_result(method, analysis_id)
-        return response.result, response.result["error"]
+        return response.result.to_dict(), response.result.error
     else:
         raise UnknownAnalysisType(
             "Cannot determine analysis type from identifier: %r", analysis_id
@@ -1159,13 +1198,38 @@ def install(
 
 
 @with_api_client
-def list_thoth_container_images(api_client: ApiClient) -> typing.Dict[str, Any]:
+def list_thoth_container_images(
+    api_client: ApiClient,
+    *,
+    os_name: typing.Optional[str] = None,
+    os_version: typing.Optional[str] = None,
+    python_version: typing.Optional[str] = None,
+    cuda_version: typing.Optional[str] = None,
+    image_name: typing.Optional[str] = None,
+    library_name: typing.Optional[str] = None,
+    symbol: typing.Optional[str] = None,
+    package_name: typing.Optional[str] = None,
+    rpm_package_name: typing.Optional[str] = None,
+) -> typing.List[typing.Dict[str, Any]]:
     """Get available Thoth container images."""
-    return (
-        ContainerImagesApi(api_client)
-        .list_thoth_container_images()
-        .to_dict()["container_images"]
-    )
+    filtering = locals()
+    filtering.pop("api_client", None)
+    filtering = {k: v for k, v in filtering.items() if v is not None}
+
+    result = []
+    page = 0
+    while True:
+        images = (
+            ContainerImagesApi(api_client)
+            .list_thoth_container_images(page=page, **filtering)
+            .to_dict()["container_images"]
+        )
+        if not images:
+            break
+        result.extend(images)
+        page += 1
+
+    return result
 
 
 @with_api_client
@@ -1186,69 +1250,104 @@ def list_python_package_indexes(api_client: ApiClient) -> typing.Dict[str, Any]:
 
 @with_api_client
 def get_package_from_imported_packages(
-    api_client: ApiClient, import_name: str
+    api_client: ApiClient,
+    import_name: str,
+    raise_on_error: bool = True,
 ) -> typing.List[Dict[str, Any]]:
     """Get all (package_name, package_version, index_url) triplets for given import package name."""
-    return (
-        PythonPackagesApi(api_client)
-        .get_package_from_imported_packages(import_name)
-        .to_dict()["package_names"]
-    )
+    try:
+        return (
+            PythonPackagesApi(api_client)
+            .get_package_from_imported_packages(import_name)
+            .to_dict()["package_names"]
+        )
+    except ApiException as exc:
+        error_message = f"Failed to obtain package for import {import_name!r} (HTTP status {exc.status})"
+        if exc.body:
+            try:
+                error_message += (
+                    f": {str(json.loads(exc.body.decode('utf-8'))['error'])}"
+                )
+            except json.decoder.JSONDecodeError:
+                _LOGGER.error(
+                    "Failed to decode backend response (HTTP status %r)", exc.status
+                )
+
+        if raise_on_error:
+            raise NoMatchingPackage(error_message) from exc
+
+        return []
 
 
-def get_verified_packages_from_static_analysis(src_path: str = "."):
+def get_verified_packages_from_static_analysis(
+    src_path: str = ".",
+    *,
+    without_standard_imports: bool = False,
+    without_builtin_imports: bool = False,
+    without_builtins: bool = False,
+    raise_on_error: bool = True,
+) -> typing.List[typing.Dict[str, str]]:
     """Get verified packages from invectio static analysis result."""
     # 1. Obtain list of imports using invectio
-    result = get_static_analysis(src_path)
-    packages = []
+    result = get_static_analysis(
+        src_path,
+        without_standard_imports=without_standard_imports,
+        without_builtin_imports=without_builtin_imports,
+        without_builtins=without_builtins,
+    )
 
+    packages = []
     if result:
         packages = [p for p in result["report"].keys()]
 
+    if not packages:
+        _LOGGER.error("No package imports identified in %r", src_path)
+        return []
+
     # 2. For each import verify package (name, version, index) (whatprovides logic)
     verified_packages = []
 
     for import_name in packages:
         unique_packages: typing.List[typing.Dict] = []
 
-        try:
-            imported_packages = get_package_from_imported_packages(import_name)
+        imported_packages = get_package_from_imported_packages(
+            import_name, raise_on_error=raise_on_error
+        )
+        if not imported_packages:
+            _LOGGER.error("No package for import %r found", import_name)
+            continue
 
-            if imported_packages:
-                for package in imported_packages:
-                    if package["package_name"] not in [
-                        p["package_name"] for p in unique_packages
-                    ]:
-                        unique_packages.append(
-                            {
-                                "package_name": package["package_name"],
-                                "index_url": package["index_url"],
-                            }
-                        )
-                    else:
-                        existing_indexes = [
-                            p["index_url"]
-                            for p in unique_packages
-                            if p["package_name"] == package["package_name"]
-                        ]
+        for package in imported_packages:
+            if package["package_name"] not in [
+                p["package_name"] for p in unique_packages
+            ]:
+                unique_packages.append(
+                    {
+                        "package_name": package["package_name"],
+                        "index_url": package["index_url"],
+                    }
+                )
+            else:
+                existing_indexes = [
+                    p["index_url"]
+                    for p in unique_packages
+                    if p["package_name"] == package["package_name"]
+                ]
 
-                        if package["index_url"] not in existing_indexes:
-                            unique_packages.append(
-                                {
-                                    "package_name": package["package_name"],
-                                    "index_url": package["index_url"],
-                                }
-                            )
-
-                for unique_package in unique_packages:
-                    verified_packages.append(unique_package)
-                    _LOGGER.info(
-                        f"Package name {unique_package['package_name']} identifed for import name {import_name}"
+                if package["index_url"] not in existing_indexes:
+                    unique_packages.append(
+                        {
+                            "package_name": package["package_name"],
+                            "index_url": package["index_url"],
+                        }
                     )
 
-        except Exception as error:
-            _LOGGER.warning(
-                f"No packages identified for import name {import_name}: {error}"
+        for unique_package in unique_packages:
+            verified_packages.append(unique_package)
+            _LOGGER.info(
+                "Package name %r identified for import %r",
+                unique_package["package_name"],
+                import_name,
             )
 
     return verified_packages
@@ -1392,7 +1491,7 @@ def collect_support_information_dict() -> Dict[str, Any]:
 
     last_analysis_id = None
     try:
-        last_analysis_id = _get_last_analysis_id()
+        last_analysis_id = get_last_analysis_id()
     except FileNotFoundError:
         _LOGGER.warning("Cannot retrieve last analysis identifier")
         pass
@@ -1523,10 +1622,13 @@ def print_dependency_graph(
     analysis_id: typing.Optional[str] = None, *, fold: bool = True
 ) -> bool:
     """Print dependency graph to stdout produced by the given adviser."""
-    if not analysis_id:
-        adviser_document, error = get_analysis_results()
-    else:
-        adviser_document, error = get_analysis_results(analysis_id)
+    analysis_id = analysis_id or get_last_analysis_id()
+    if not analysis_id.startswith("adviser-"):
+        raise UnknownAnalysisType(
+            f"An adviser identifier is required to print a dependency graph, got {analysis_id!r} instead"
+        )
+
+    adviser_document, error = get_analysis_results(analysis_id)
 
     if error:
         print("Cannot print dependency graph as advise was not successful")
@@ -1565,3 +1667,105 @@ def add_requirements_to_project(
         "Changes done might require triggering new advise to resolve dependencies"
     )
     thoth_config.save_project(project)
+
+
+def check_runtime_environment_run(runtime_environment: Dict[str, Any]) -> None:
+    """Check the runtime environment matches runtime environment declared in the config file."""
+    discovered = discover_all()
+
+    if runtime_environment.get("base_image") != discovered.get("base_image"):
+        raise PedanticRunVerificationError(
+            f"Base image configured {runtime_environment.get('base_image')!r} does not match "
+            f"base image discovered {discovered.get('base_image')!r}"
+        )
+
+    if runtime_environment.get("cuda_version") != discovered.get("cuda_version"):
+        raise PedanticRunVerificationError(
+            f"CUDA version configured {runtime_environment.get('cuda_version')!r} does not match "
+            f"CUDA version discovered {discovered.get('cuda_version')!r}"
+        )
+
+    if runtime_environment.get("platform", "linux-x86_64") != discovered.get(
+        "platform"
+    ):
+        raise PedanticRunVerificationError(
+            f"Platform configured {runtime_environment.get('platform')!r} does not match "
+            f"platform discovered {discovered.get('platform')!r}"
+        )
+
+    if runtime_environment.get("python_version") != discovered.get("python_version"):
+        raise PedanticRunVerificationError(
+            f"Python version configured {runtime_environment.get('python_version')!r} does not match "
+            f"Python version discovered {discovered.get('python_version')!r}"
+        )
+
+    hw_discovered = discovered.get("hardware") or {}
+    hw_configured = runtime_environment.get("hardware") or {}
+    if hw_configured.get("cpu_family") is not None and hw_configured.get(
+        "cpu_family"
+    ) != hw_discovered.get("cpu_family"):
+        raise PedanticRunVerificationError(
+            f"CPU family configured {hw_configured.get('cpu_family')!r} does not match "
+            f"CPU family discovered {hw_discovered.get('cpu_family')!r}"
+        )
+
+    if hw_configured.get("cpu_model") is not None and hw_configured.get(
+        "cpu_model"
+    ) != hw_discovered.get("cpu_model"):
+        raise PedanticRunVerificationError(
+            f"CPU model configured {hw_configured.get('cpu_model')!r} does not match "
+            f"CPU model discovered {hw_discovered.get('cpu_model')!r}"
+        )
+
+    if hw_configured.get("gpu_model") is not None and hw_configured.get(
+        "gpu_model"
+    ) != hw_discovered.get("gpu_model"):
+        raise PedanticRunVerificationError(
+            f"GPU model configured {hw_configured.get('gpu_model')!r} does not match "
+            f"GPU model discovered {hw_discovered.get('gpu_model')!r}"
+        )
+
+    os_discovered = discovered.get("operating_system") or {}
+    os_configured = runtime_environment.get("operating_system") or {}
+    if os_configured.get("name") != os_discovered.get("name"):
+        raise PedanticRunVerificationError(
+            f"Operating system name configured {os_configured.get('name')!r} does not match "
+            f"operating system name discovered {os_discovered.get('name')!r}"
+        )
+
+    if os_configured.get("version") != os_discovered.get("version"):
+        raise PedanticRunVerificationError(
+            f"Operating system version configured {os_configured.get('version')!r} does not match "
+            f"operating system version discovered {os_discovered.get('version')!r}"
+        )
+
+    _LOGGER.info("Runtime environment configured matches runtime environment used")
+
+
+def load_dot_env(dot_env_path: str) -> Dict[str, str]:
+    """Load .env file and parse its content."""
+    if not os.path.isfile(dot_env_path):
+        _LOGGER.debug("No .env file at %r found", dot_env_path)
+        return {}
+
+    _LOGGER.info("Loading environment variables from %r", dot_env_path)
+
+    env = {}
+    with open(dot_env_path) as f:
+        for line_no, line in enumerate(f.readlines()):
+            line = line.strip()
+            if not line or line.startswith("#"):  # Skip comments.
+                continue
+
+            parts = line.split("=", maxsplit=1)
+            if len(parts) != 2:
+                _LOGGER.error(
+                    "Failed to parse %r content at line %d, ignoring...",
+                    dot_env_path,
+                    line_no,
+                )
+                continue
+
+            env[parts[0]] = parts[1]
+
+    return env
